import gtk
import gobject
import cairo
import math

class flipview(gtk.DrawingArea):
    """this class creates a cover flow like widget from an array of images
    with cairo"""
    
    _BUTTONSIZE = 40
    
    _CAIRO = False
    _OBJECTS = []
    _WIDTH = 0
    _PAD = 0
    _SCALE = 0
    _SPEED = 0
    _SHIFT = 0
    _SHOW = 0
    _POS = 0.0
    _TARG = 0.0
    _DIR = 1
    _RUNNING = False
    _LEFTCLICK = False
    _RIGHTCLICK = False
    _DOUBLECLICK = False
    _MIDDLECLICK = False
    _OVER = 0
    _INTERPOLATION = gtk.gdk.INTERP_NEAREST
    _COLOR = False
    
    __gsignals__ = { "expose-event": "override" }
    
    def __init__(self, objects, vpos, width, padding, scale, interpolation, show,
                       start, speed, background, color, leftClick, rightClick, middleClick,
                       doubleClick):
        super(flipview,self).__init__()
        self._OBJECTS = objects
        self._WIDTH = width
        self._PAD = padding
        self._SCALE = scale
        self._SPEED = speed
        self._VPOS = vpos
        self._SHOW = show
        self._COLOR = color
        self._BACKGROUND = background
        self._INTERPOLATION = interpolation
        if self._SHOW % 2 == 0:
             self._SHOW += 1
        self._LEFTCLICK = leftClick
        self._RIGHTCLICK = rightClick
        self._MIDDLECLICK = middleClick
        self._DOUBLECLICK = doubleClick
        self._POS = float(start)
        self._TARG = float(start)
        self._initObjects(self._OBJECTS)
        self.add_events( gtk.gdk.BUTTON_PRESS_MASK
                       | gtk.gdk.BUTTON_RELEASE_MASK
                       | gtk.gdk.SCROLL_MASK
                       | gtk.gdk.POINTER_MOTION_MASK
                       | gtk.gdk.POINTER_MOTION_HINT_MASK
                       | gtk.gdk.ENTER_NOTIFY_MASK
                       | gtk.gdk.LEAVE_NOTIFY_MASK)
        self.connect("motion_notify_event", self.mouseMove)
        self.connect("enter_notify_event", self.mouseOver)
        self.connect("leave_notify_event", self.mouseOut)
        self.connect("button_press_event", self.click)
        self.connect("scroll_event", self.scroll)
    
    # public stuff
    def redraw(self):
        self.queue_draw()
    
    def trigger(self, i):
        if i < 0: i = 0
        if i >= len(self._OBJECTS):
            i = len(self._OBJECTS) - 1
        self._TARG = float(i)
        if not self._RUNNING:
            self._run()
    
    def getActive(self, i):
        return self._OBJECTS[i]
    
    def addObject(self, obj, pos):
        """add a new object to the list of objects and create basic stuff"""
        try:
            if pos < 0: pos = count(self._OBJECTS) - pos
            if pos > count(self._OBJECTS): pos = count(self._OBJECTS)
        except:
            pos = count(self._OBJECTS)
        if not obj["pixbuf"]:
            obj["pixbuf"] = gtk.gdk.pixbuf_new_from_file(obj["image"])
        self._OBJECTS.insert(pos, obj)
        self.trigger(pos)
    
    def setActive(self, i):
        try:
            self._OBJECTS[i]["active"] = True
            self.trigger(i)
        except:
            pass
            
    def setInactive(self, i):
        if i < 0:
            for o in self._OBJECTS:
                o["active"] = False
        else:
            try:
                self._OBJECTS[i]["active"] = False
            except:
                pass
            
    # signals & events
    def click(self, widget, event):
        x = event.x
        y = event.y
        w = float(self.get_allocation().width)
        if (x > w - (w * (self._WIDTH / 100.0)) / 2
        or x < (w * (self._WIDTH / 100.0)) / 2) and event.button == 1:
            # left click on arrows
            self.trigger(self._TARG + self._OVER)
        elif event.type == gtk.gdk._2BUTTON_PRESS and self._DOUBLECLICK:
            # double click
            self._DOUBLECLICK(int(self._TARG), event)
        elif event.button == 1 and self._LEFTCLICK:
            self._LEFTCLICK(int(self._TARG), event)
        elif event.button == 2 and self._MIDDLECLICK:
            self._MIDDLECLICK(int(self._TARG), event)
        elif event.button == 3 and self._RIGHTCLICK:
            self._RIGHTCLICK(int(self._TARG), event)
                
    def scroll(self, widget, event):
        if (event.direction == gtk.gdk.SCROLL_UP
        or event.direction == gtk.gdk.SCROLL_LEFT):
            self.trigger(self._TARG - 1)
        elif (event.direction == gtk.gdk.SCROLL_DOWN
        or event.direction == gtk.gdk.SCROLL_RIGHT):
            self.trigger(self._TARG + 1)
        
    def mouseOver(self, widget, event):
        pass
        
    def mouseOut(self, widget, event):
        self._OVER = 0
        self.redraw()
        
    def mouseMove(self, widget, event):
        # if this is a hint, then let's get all the necessary
        # information, if not it's all we need.
        if event.is_hint:
            x, y, state = event.window.get_pointer()
        else:
            x = event.x
            y = event.y
            state = event.state
        w = float(self.get_allocation().width)
        if x > w - (w * (self._WIDTH / 100.0)) / 2:
            o = 1
        elif x < (w * (self._WIDTH / 100.0)) / 2:
            o = -1
        else:
            o = 0
        if(o != self._OVER):
            self._OVER = o
            self.redraw()
    
    def _initObjects(self, objects):
        """Internal handler to create pixpufs for all objects"""
        for o in objects:
            o["pixbuf"] = gtk.gdk.pixbuf_new_from_file(o["image"])
        
    # motion & drawing
    def _run(self, blsa=False):
        # increase / decrease position
        if self._POS != self._TARG:
            self._POS += (self._TARG - self._POS) / 2 * self._SPEED
        # denormals
        if self._POS < self._TARG + 0.005 and self._POS > self._TARG - 0.005:
            self._POS = self._TARG
        # switch dir if above limits
        if self._POS > len(self._OBJECTS) - 1:
            self._POS = len(self._OBJECTS) - 1
            self._TARG = self._POS
            self.switchDir()
        if self._POS < 0:
            self._POS = 0
            self._TARG = self._POS
            self.switchDir()
        # draw to canvas
        self.redraw()
        # call next run
        if self._POS != self._TARG:
            gobject.timeout_add(40, self._run, self)
            self._RUNNING = True
        else:
            self._RUNNING = False
    
    def draw(self):
        rect = self.get_allocation()
        
        canvasw = rect.width
        canvash = rect.height
        
        padx = canvasw * (self._PAD / 100.0)
        pady = canvash * (self._PAD / 100.0)
        
        draww = canvasw - 2 * padx
        drawh = canvash - 2 * pady
        
        imgmaxw = canvasw * (self._WIDTH / 100.0)
        imgminw = imgmaxw * (self._SCALE / 100.0)
        imgmaxh = canvash - pady * 2
        imgminh = imgmaxh * (self._SCALE / 100.0)
        
        surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, canvasw, canvash)
        cr = cairo.Context(surface)
        
        # background
        r, g, b = self._BACKGROUND
        cr.rectangle(0, 0, canvasw, canvash)
        cr.set_source_rgb(r, g, b)
        
        cr.fill()
        start = int(round(self._POS - (self._SHOW - 1) / 2))
        stop = int(round(self._POS + (self._SHOW - 1) / 2) + 1)
        for i in range(start, stop):
            # counter starting @ 0
            c = i - int(round(self._POS - (self._SHOW - 1) / 2))
            # reorder drawing queue
            if c % 2 == 1:
                p = self._SHOW - int(c / 2) - 1
            else:
                p = c / 2
            # obkject id
            p += start
            if p >= 0 and p < len(self._OBJECTS):
                obj = self._OBJECTS[p]
                # calculate some stuff
                # get linear value between -1 and 1 depending on position
                pos = (p - self._POS) / ((self._SHOW - 1) / 2);
                pos = min(1, pos)
                pos = max(-1, pos)
                # raise to value between 0 and 1
                xpos = (pos + 1) * 0.5;
                # calculate grad between 0 and 180
                angle = xpos * 180;
                
                # calculate radian measure
                radpos = (math.pi / 180) * angle
                # calculate x-coeff position (from -1 to 1)
                xpcoeff = math.cos(radpos) * -1;
                # switch to value between 0 and 1
                xpcoeff1 = (xpcoeff + 1) * 0.5;
                # get a value between 0, 1 and 0
                xpc1 = xpcoeff
                if xpc1 < 0: xpc1 *= -1
                size = 1 - xpc1
                #psize = (self._SCALE + size * (100 - self._SCALE)) / 100
                
                # calculate radian measure
                radsize = (math.pi / 180) * (angle)
                # calculate x-coeff size (from 0 to 1 to 0)
                xscoeff = math.sin(radsize)
                # calc size value in percent
                psize = (self._SCALE + xscoeff * (100 - self._SCALE)) / 100
                
                # original size
                imgw = float(obj["pixbuf"].get_width())
                imgh = float(obj["pixbuf"].get_height())
                
                # target max
                targmaxw = imgmaxw * psize
                targmaxh = imgmaxh * psize
                
                # ratio
                ratiow = imgw / targmaxw
                ratioh = imgh / targmaxh
                
                # target size
                if (ratiow < ratioh):
                    targw = imgw * (1 / ratioh)
                    targh = imgh * (1 / ratioh)
                else:
                    targw = imgw * (1 / ratiow)
                    targh = imgh * (1 / ratiow)
                
                # scale width
                #targw = max(targw*size, 1.0)
                
                # position
                maxab = draww - imgminw
                xcenter = (xpcoeff1 * maxab) + padx + (imgminw / 2)
                imgx = xcenter - (targw / 2)
                imgy = (drawh - targh) * self._VPOS + pady

                # draw image
                ct = gtk.gdk.CairoContext(cr)
                ct.set_source_pixbuf(obj["pixbuf"].scale_simple(int(targw),
                                                      int(targh),
                                                      self._INTERPOLATION),
                                                      imgx, imgy)
                ct.paint_with_alpha(size)
                
                # marker
                try:
                    if obj["active"]:
                        ct.rectangle(imgx + 3,
                                     imgy + 3,
                                     int(targw) - 6,
                                     int(targh) - 6)
                        ct.set_line_width(6)
                        r, g, b, a = self._COLOR
                        ct.set_source_rgba(r, g, b, size * a);
                        ct.stroke()
                        
                except:
                    pass
                
                # buttons
                ct.set_line_width(2)
                if self._OVER > 0:
                    x0 = canvasw - self._BUTTONSIZE * 2
                    x1 = canvasw - self._BUTTONSIZE * 2
                    xc = canvasw - self._BUTTONSIZE
                    r = self._BUTTONSIZE / 2
                    cr.rectangle(x0, 0, x1, canvash)
                    cr.set_source_rgba(0, 0, 0, 0.5);
                    cr.fill()
                    
                    cr.move_to(xc - r, canvash / 2 - r)
                    cr.line_to(xc + r, canvash / 2)
                    cr.line_to(xc - r, canvash / 2 + r)
                    cr.line_to(xc - r, canvash / 2 - r)
                    cr.set_source_rgb(1, 1, 1);
                    cr.stroke()
                if self._OVER < 0:
                    x0 = 0
                    x1 = self._BUTTONSIZE * 2
                    xc = self._BUTTONSIZE
                    r = self._BUTTONSIZE / 2
                    cr.rectangle(x0, 0, x1, canvash)
                    cr.set_source_rgba(0, 0, 0, 0.5);
                    cr.fill()
                    
                    cr.move_to(xc + r, canvash / 2 - r)
                    cr.line_to(xc + r, canvash / 2 + r)
                    cr.line_to(xc - r, canvash / 2)
                    cr.line_to(xc + r, canvash / 2 - r)
                    cr.set_source_rgb(1, 1, 1);
                    cr.stroke()
        # text
        try:
            text = self._OBJECTS[int(self._TARG)]["title"]
            cr.select_font_face('Sans', cairo.FONT_SLANT_NORMAL,
                                        cairo.FONT_WEIGHT_NORMAL)
            cr.set_font_size(15)
            fascent, fdescent, fheight, fxadvance, fyadvance = cr.font_extents()
            xbearing, ybearing, width, height, xadvance, yadvance = \
                    cr.text_extents(text)
            tx = canvasw / 2 - xbearing - width / 2
            ty = canvash - fdescent - padx
            
            cr.rectangle(tx - 15, ty - 10 - fheight + fdescent,
                         xbearing + width + 30,
                         xbearing + fheight + 20)
            cr.set_source_rgba(0, 0, 0, 0.8);
            cr.fill()
                    
            cr.move_to(tx, ty)
            cr.set_source_rgb(1, 1, 1)
            cr.show_text(text)
        except:
            pass
                
        self._CAIRO.set_source_surface(surface)
        self._CAIRO.paint()
        return True
    
    # little helpers
    def do_expose_event(self, event):
        self._CAIRO = self.window.cairo_create()
        self._CAIRO.rectangle(event.area.x, event.area.y,
                     event.area.width, event.area.height)
        self._CAIRO.clip()
        self.draw()
        
    def switchDir(self):
        self._DIR *= -1
